"""eepackages - A set of utilities built on top of Google Earth Engine (migrated from JavaScript)"""

__version__ = "0.1.0"
__author__ = "Gennadii Donchyts <gennadiy.donchyts@gmail.com>"
__all__ = []
